/* Distributed Checksum Clearinghouse
 *
 * compute fuzzy body checksum #2
 *
 * Copyright (c) 2003 by Rhyolite Software
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND RHYOLITE SOFTWARE DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL RHYOLITE SOFTWARE
 * BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Rhyolite Software DCC 1.2.28-1.28 $Revision$
 */

#include "ck.h"
#include "ckfuz2_tbl.h"

#define FZ2 cks->fuz2


void
dcc_ck_fuz2_init(DCC_GOT_CKS *cks)
{
	FZ2.wlen = 0;
	DCC_FUZ2_WORD_CLEAR(&FZ2.w);
	FZ2.state = DCC_FUZ2_ST_WORD;
	FZ2.url_cp = FZ2.url_buf;
	FZ2.urls = 0;

	cks->sums[DCC_CK_FUZ2].type = DCC_CK_FUZ2;
	cks->sums[DCC_CK_FUZ2].tgts = DCC_TGTS_INVALID;
	cks->sums[DCC_CK_FUZ2].rpt = 0;

	FZ2.btotal = 0;
	FZ2.summed = 0;
	FZ2.wtotal = 0;

	MD5Init(&FZ2.md5);
}



static u_char				/* 1=found it, 0=not a known word */
lookup_word(const DCC_FUZ2_WORD *w, u_int wlen)
{
	const char *p;
	u_int n;

	p = word_tbl[fuz2_word_hash(w, WORD_HASH_LEN)];
	if (!p)
		return 0;
	for (;;) {
		n = *p++;
		if (!n)
			return 0;
		if (n == wlen && !memcmp(w->b, p, n))
			return 1;
		p += n;
	}
}



/* resolve an HTML characater reference */
static u_char				/* 0 or found value */
lookup_cref(DCC_FUZ2_WORD *c, u_int clen)
{
	const char *p;
	u_int n;

	if (clen > sizeof(DCC_FUZ2_WORD)
	    || clen == 0)
		return 0;
	p = cref_tbl[fuz2_word_hash(c, CREF_HASH_LEN)];
	if (!p)
		return 0;
	for (;;) {
		n = *p++;
		if (!n)
			return 0;
		if (n == clen && !memcmp(c->b, p, n))
			return p[clen];
		p += n+1;
	}
}



/* add a character reference to the growing word or end the word */
static u_char				/* 0=don't consume end of char ref */
add_cref(DCC_GOT_CKS *cks,
	 u_char end_cref,		/* what ended the cref */
	 u_char cref)
{
#define SKIP_WORD() (FZ2.wlen = sizeof(DCC_FUZ2_WORD)+1)

	FZ2.state = DCC_FUZ2_ST_WORD;
	cref = fc[cref];
	if (cref >= FC_A) {
		if (FZ2.wlen >= sizeof(DCC_FUZ2_WORD)) {
			SKIP_WORD();
		} else {
			FZ2.w.b[FZ2.wlen++] = cref;
		}
	} else if (cref == FC_SK) {
		;			/* ignore accent marks */
	} else if (cref == FC_SP || cref == FC_CF) {
		/* &nbsp simply ends words
		 * use end of cref to end the word */
		return 0;
	} else {
		SKIP_WORD();		/* end word at junk */
	}

	/* end word if cref had no ';' */
	return (end_cref == ';');
#undef SKIP_WORD
}



void
dcc_ck_fuz2(DCC_GOT_CKS *cks, const char *bp, u_int bp_len)
{
#define SKIP_WORD() (FZ2.wlen = sizeof(DCC_FUZ2_WORD)+1)
#define JUNK() (SKIP_WORD(), FZ2.state = DCC_FUZ2_ST_WORD)
#	define BUF_LEN 1024
	char buf[BUF_LEN+sizeof(DCC_FUZ2_WORD)+1];
	int blen;
	int i;
	char *p;
	u_char c;

	if (cks->sums[DCC_CK_FUZ2].type != DCC_CK_FUZ2)
		return;

	blen = 0;
	while (bp_len != 0) {
		switch (FZ2.state) {
		case DCC_FUZ2_ST_WORD:
			/* gathering a word */
			do {
				--bp_len;
				c = fc[(u_char)*bp++];
				if (c >= FC_A) {
					++FZ2.btotal;
					if (FZ2.wlen < sizeof(DCC_FUZ2_WORD))
					    FZ2.w.b[FZ2.wlen++] = c;
					else
					    SKIP_WORD();
					continue;
				}

				if (c == FC_CF) {
					if (cks->mime_ct == DCC_CK_CT_HTML) {
					    FZ2.clen = 0;
					    DCC_FUZ2_WORD_CLEAR(&FZ2.c);
					    FZ2.state = DCC_FUZ2_ST_CREF;
					    break;
					}
					c = FC_SP;
				}

				if (c == FC_SP) {
					if (FZ2.wlen >= MIN_WLEN
					    && FZ2.wlen <= sizeof(DCC_FUZ2_WORD)
					    && lookup_word(&FZ2.w, FZ2.wlen)) {
					    ++FZ2.wtotal;
					    memcpy(&buf[blen],&FZ2.w, FZ2.wlen);
					    blen += FZ2.wlen;
					    if (blen >= BUF_LEN) {
						FZ2.summed += blen;
						MD5Update(&FZ2.md5, buf, blen);
						blen = 0;
					    }
					}
					FZ2.wlen = 0;
					DCC_FUZ2_WORD_CLEAR(&FZ2.w);
					continue;
				}
				++FZ2.btotal;

				if (c == FC_LT) {
					FZ2.tlen = 0;
					DCC_FUZ2_WORD_CLEAR(&FZ2.c);
					FZ2.state = DCC_FUZ2_ST_START_TAG;
					break;
				}

				JUNK();
			} while (bp_len != 0);
			break;

		case DCC_FUZ2_ST_CREF:
			/* start to get a possible HTML character reference
			 * We've passed the '&' */
			c = *bp;
			if (c == '#') {
				FZ2.state = DCC_FUZ2_ST_CREF_NUM;
			} else if ((c >= 'a' && c <= 'z')
				   || (c >= 'A' && c <= 'Z')) {
				FZ2.c.b[0] = c;
				FZ2.clen = 1;
				FZ2.state = DCC_FUZ2_ST_CREF_NAME;
			} else {
				JUNK();
				break;	/* check this character again */
			}
			++FZ2.btotal;
			++bp, --bp_len;
			break;

		case DCC_FUZ2_ST_CREF_NUM:
			/* look for 'x' or the first digit */
			c = *bp;
			if (c == 'x' || c == 'X') {
				FZ2.state = DCC_FUZ2_ST_CREF_HEX;
			} else if (c >= '0' && c <= '9') {
				FZ2.clen = c - '0';
				FZ2.state = DCC_FUZ2_ST_CREF_DEC;
			} else {
				JUNK();
				break;
			}
			++FZ2.btotal;
			++bp, --bp_len;
			break;

		case DCC_FUZ2_ST_CREF_DEC:
			c = *bp;
			if (c >= '0' && c <= '9') {
				FZ2.clen = FZ2.clen*10 + (c - '0');
			} else if (fc[c] == FC_SP) {
				if (!add_cref(cks, c, FZ2.clen))
					break;
				/* sometimes check this character again */
			} else {
				JUNK();
			}
			if (FZ2.clen > 256)
				FZ2.clen = 256;
			++FZ2.btotal;
			++bp, --bp_len;
			break;

		case DCC_FUZ2_ST_CREF_HEX:
			c = *bp;
			if ((c >= 'a' && c <= 'f')
			    || (c >= 'A' && c <= 'F')) {
				FZ2.clen = (FZ2.clen<<4) + (c & 0xf) + 9;
			} else if (c >= '0' && c <= '9') {
				FZ2.clen = (FZ2.clen<<4) + (c - '0');
			} else if (fc[c] == FC_SP) {
				if (!add_cref(cks, c, FZ2.clen))
					break;
				/* sometimes check this character again */
			} else {
				JUNK();
			}
			if (FZ2.clen > 256)
				FZ2.clen = 256;
			++FZ2.btotal;
			++bp, --bp_len;
			break;

		case DCC_FUZ2_ST_CREF_NAME:
			c = *bp;
			if ((c >= 'a' && c <= 'z')
			    || (c >= 'A' && c <= 'Z')) {
				/* If the word is too long, we won't match it.
				 * This will consume it */
				if (FZ2.clen < sizeof(DCC_FUZ2_WORD))
					FZ2.c.b[FZ2.clen++] = c;
			} else if (fc[c] == FC_SP) {
				/* this character ends the cref */
				if (!add_cref(cks, c, lookup_cref(&FZ2.c,
							FZ2.clen)))
					break;	/* consume char ending cref */
				/* treat ending char separately */
			} else {
				JUNK();
			}
			++FZ2.btotal;
			++bp, --bp_len;
			break;

		case DCC_FUZ2_ST_START_TAG:
			/* collecting an HTML tag or comment
			 * We've passed the '<' */
			c = *bp;
#define SAVE_TAG(_c) (FZ2.c.b[FZ2.tlen++] = _c, ++FZ2.btotal, ++bp, --bp_len)
			if (((c >= 'a' && c <= 'z')	/* tag */
			     || (c >= '0' && c <= '9'))
			    && FZ2.tlen < sizeof(DCC_FUZ2_WORD)) {
				SAVE_TAG(c);
				break;
			}
			if (c >= 'A' && c <= 'Z'
			    && FZ2.tlen < sizeof(DCC_FUZ2_WORD)) {
				SAVE_TAG(c - ('A'-'a'));
				break;
			}
			if ((c == '/'	/* end-tag */
			     || c == '!')   /* start of comment */
			    && FZ2.tlen == 0) {
				SAVE_TAG(c);
				break;
			}
			if (c == '-'    /* comment */
			    && FZ2.tlen >= 1 && FZ2.tlen <= 2) {
				SAVE_TAG(c);
				break;
			}
#undef SAVE_TAG
			if (FZ2.tlen == 4
			    && cks->mime_ct != DCC_CK_CT_HTML
			    && !memcmp(FZ2.c.b, "html", 4))
				cks->mime_ct = DCC_CK_CT_HTML;

			if (cks->mime_ct == DCC_CK_CT_HTML
			    && FZ2.tlen > 0) {
				/* if we are in an HTML document and we
				 * have at least one character after '<',
				 * assume it is some kind of HTML tag */
				FZ2.summed += FZ2.tlen+1;   /* count '<' */
				if (c == '>') {
					/* optimize common simple tags */
					++FZ2.summed;
					++FZ2.btotal;
					++bp, --bp_len;
					FZ2.state = DCC_FUZ2_ST_WORD;
					break;
				}
				if (FZ2.tlen >= 3
				    && !memcmp(FZ2.c.b, "!--", 3)) {
					FZ2.state = DCC_FUZ2_ST_SKIP_COMMENT;
				} else {
					FZ2.url.st = DCC_URL_ST_IDLE;
					FZ2.state = DCC_FUZ2_ST_SKIP_TAG;
				}
			} else {
				/* assume it is not an HTML tag and
				 * mark the whole word as junk */
				JUNK();
			}
			break;

		case DCC_FUZ2_ST_SKIP_TAG:
			/* skip rest of boring HTML tag
			 * ought to ignore '>' in quotes */
			do {
				--bp_len;
				c = *bp++;

				/* capture URLs */
				i = dcc_ck_url(&FZ2.url, c, &FZ2.url_cp);
				c = i & ~DCC_CK_URL_MASK;
				switch (i - c) {
				case DCC_CK_URL_CHAR:
					break;
				case DCC_CK_URL_CK_LEN:
					/* Make room if we are too close to end
					 * of the buffer for maximum size URL */
					while (FZ2.url_cp
					       >= &FZ2.url_buf[ISZ(FZ2.url_buf)
							- DCC_FUZ2_URL_MAX]) {
					    p = memchr(FZ2.url_buf, '\0',
						       FZ2.url_cp-FZ2.url_buf);
					    ++p;
					    memmove(FZ2.url_buf, p,
						    FZ2.url_cp - p);
					    FZ2.url_cp -= p - FZ2.url_buf;
					}
					if (FZ2.url_cp != FZ2.url_buf)
					    *FZ2.url_cp++ = '\0';
					++FZ2.urls;
					break;
				case DCC_CK_URL_HOST:
					*FZ2.url_cp++ = c;
					continue;
				case DCC_CK_URL_SKIP:
					continue;
				}

				if (c == '>') {
					++FZ2.summed;
					++FZ2.btotal;
					FZ2.state = DCC_FUZ2_ST_WORD;
					break;
				}
				if (fc[c] != FC_SP) {
					++FZ2.summed;
					++FZ2.btotal;
					/* don't let wild tags run forever */
					if (++FZ2.tlen > DCC_URL_FAILSAFE) {
					    JUNK();
					    break;
					}
				}
			} while (bp_len != 0);
			break;

		case DCC_FUZ2_ST_SKIP_COMMENT:
			/* HTML comments can include HTML tags,
			 * but spammers don't understand HTML comment syntax
			 * and Netscape and IE treat (and ignore) broken
			 * comments like strange tags. */
			do {
				--bp_len;
				c = *bp++;
				if (c == '>') {
					++FZ2.summed;
					++FZ2.btotal;
					FZ2.state = DCC_FUZ2_ST_WORD;
					break;
				}
				if (fc[c] != FC_SP) {
					++FZ2.summed;
					++FZ2.btotal;
					/* don't let wild tags run forever */
					if (++FZ2.tlen > DCC_URL_FAILSAFE) {
					    JUNK();
					    break;
					}
				}
			} while (bp_len != 0);
			break;
		}
	}
	if (blen != 0) {
		FZ2.summed += blen;
		MD5Update(&FZ2.md5, buf, blen);
	}
#undef SKIP_WORD
#undef JUNK
#undef BUF_LEN
}



void
dcc_ck_fuz2_fin(DCC_GOT_CKS *cks)
{
	if (cks->sums[DCC_CK_FUZ2].type != DCC_CK_FUZ2)
		return;

#ifdef DCC_CKSUM_DEBUG
	if (dcc_clnt_debug > 3)
		printf("\n***fuz2: wtotal=%d summed=%d btotal=%d\n",
		       FZ2.wtotal, FZ2.summed, FZ2.btotal);
#endif
	/* The FUZ2 checksum is not valid if it is on a few words and
	 * less than 10% of a big, binary file */
	if (FZ2.wtotal < 100
	    && FZ2.summed*10 < FZ2.btotal) {
		cks->sums[DCC_CK_FUZ2].type = DCC_CK_INVALID;
		return;
	}
	/* We cannot compute a checksum on a nearly empty message */
	if (FZ2.wtotal < 8) {
		if (FZ2.wtotal + FZ2.urls*4 >= 8) {
			/* use URLs if we lack words */
			MD5Update(&FZ2.md5, FZ2.url_buf,
				FZ2.url_cp - FZ2.url_buf);
		} else {
			/* Compute a checksum for local blacklising on messages
			 * that look empty to the FUZ2 checksum but are not and
			 * are not too binary. The resulting checksum is zero.
			 * Do not report it to the DCC server. */
			if (FZ2.summed >= 120) {
				memset(cks->sums[DCC_CK_FUZ2].sum, 0,
				       sizeof(cks->sums[DCC_CK_FUZ2].sum));
				cks->flags |= DCC_CKS_HAVE_SUM;
				return;
			}
			cks->sums[DCC_CK_FUZ2].type = DCC_CK_INVALID;
			return;
		}
	}

	MD5Final(cks->sums[DCC_CK_FUZ2].sum, &FZ2.md5);
	cks->sums[DCC_CK_FUZ2].rpt = 1;
	cks->flags |= DCC_CKS_HAVE_SUM;
}
